@@ -22,7 +22,7 @@
  * @author    Tobias Schlitt <toby@php.net>
  * @copyright 1997-2008 The PHP Group
  * @license   http://www.php.net/license/3_0.txt  PHP License 3.0
- * @version   CVS: $Id: AllTests.php,v 1.2.2.3 2008/05/19 18:08:23 jschippers Exp $
+ * @version   CVS: $Id$
  * @link      http://pear.php.net/package/Net_FTP
  * @link      http://www.phpunit.de PHPUnit
  * @since     File available since Release 0.0.1
@@ -32,10 +32,8 @@ if (!defined('PHPUnit_MAIN_METHOD')) {
     define('PHPUnit_MAIN_METHOD', 'Net_FTP_AllTests::main');
 }
 
-require_once 'PHPUnit/Framework/TestSuite.php';
 require_once 'PHPUnit/TextUI/TestRunner.php';
 
-chdir(dirname(__FILE__));
 require_once 'Net_FTPTest.php';
 
 /**
@@ -46,7 +44,7 @@ require_once 'Net_FTPTest.php';
  * @author    Jorrit Schippers <jschippers@php.net>
  * @copyright 1997-2008 The PHP Group
  * @license   http://www.php.net/license/3_0.txt  PHP License 3.0
- * @version   Release: 1.3.7
+ * @version   Release: 1.4.0
  * @link      http://pear.php.net/package/Net_FTP
  * @since     Class available since Release 1.3.3
  */
@@ -79,4 +77,4 @@ class Net_FTP_AllTests
 if (PHPUnit_MAIN_METHOD == 'Net_FTP_AllTests::main') {
     Net_FTP_AllTests::main();
 }
-?>
\ No newline at end of file
+?>
