@@ -26,7 +26,7 @@
  * @author    Tobias Schlitt <toby@php.net>
  * @copyright 1997-2008 The PHP Group
  * @license   http://www.php.net/license/3_0.txt  PHP License 3.0
- * @version   CVS: $Id: Net_FTPTest.php,v 1.7.2.5 2008/05/19 18:08:23 jschippers Exp $
+ * @version   CVS: $Id$
  * @link      http://pear.php.net/package/Net_FTP
  * @link      http://www.phpunit.de PHPUnit
  * @since     File available since Release 1.3.3
@@ -36,7 +36,6 @@ if (!defined('PHPUnit_MAIN_METHOD')) {
     define('PHPUnit_MAIN_METHOD', 'Net_FTPTest::main');
 }
 
-require_once 'PHPUnit/Framework.php';
 require_once 'System.php';
 
 chdir(dirname(__FILE__));
@@ -54,7 +53,7 @@ if (substr(dirname(__FILE__), -6) == DIRECTORY_SEPARATOR.'tests') {
  * @author    Jorrit Schippers <jschippers@php.net>
  * @copyright 1997-2008 The PHP Group
  * @license   http://www.php.net/license/3_0.txt  PHP License 3.0
- * @version   Release: 1.3.7
+ * @version   Release: 1.4.0
  * @link      http://pear.php.net/package/Net_FTP
  * @since     Class available since Release 1.3.3
  */
@@ -62,7 +61,6 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
 {
     protected $ftp;
     protected $ftpdir;
-    protected $setupError;
     
     /**
      * Runs the test methods of this class.
@@ -76,7 +74,7 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
         include_once 'PHPUnit/TextUI/TestRunner.php';
 
         $suite  = new PHPUnit_Framework_TestSuite('Net_FTPTest');
-        $result = PHPUnit_TextUI_TestRunner::run($suite);
+        PHPUnit_TextUI_TestRunner::run($suite);
     }
 
     /**
@@ -89,51 +87,36 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
     protected function setUp()
     {
         if (!file_exists('config.php')) {
-            $this->setupError = 'config.php does not exist in '.getcwd();
-            return;
+            $this->markTestSkipped('config.php does not exist in '.getcwd());
         }
         
         include_once 'config.php';
         
         if (!defined('FTPHOST') || !defined('FTPPORT') || !defined('FTPUSER')
             || !defined('FTPPASSWORD')) {
-            $this->setupError = 'Some required constants are not defined';
-            return;
+            $this->markTestSkipped('Some required constants are not defined');
         }
         
         $this->ftp = new Net_FTP(FTPHOST, FTPPORT, 30);
-        $res       = $this->ftp->connect();
-        
-        if (PEAR::isError($res)) {
-            $this->setupError = 'Could not connect to the FTP server';
-            $this->ftp        = null;
-            return;
-        }
         
         $res = $this->ftp->login(FTPUSER, FTPPASSWORD);
         
         if (PEAR::isError($res)) {
-            $this->setupError = 'Could not login to the FTP server';
-            $this->ftp        = null;
-            return;
+            $this->markTestSkipped('Could not login to the FTP server');
         }
         
         if (defined('FTPDIR') && '' !== FTPDIR) {
             $res = $this->ftp->cd(FTPDIR);
             
             if (PEAR::isError($res)) {
-                $this->setupError = 'Could switch to directory '.FTPDIR;
-                $this->ftp        = null;
-                return;
+                $this->markTestSkipped('Could switch to directory '.FTPDIR);
             }
         }
         
         $res = $this->ftp->pwd();
         
         if (PEAR::isError($res)) {
-            $this->setupError = 'Could not get current directory';
-            $this->ftp        = null;
-            return;
+            $this->markTestSkipped('Could not get current directory');
         }
         
         $this->ftpdir = $res;
@@ -141,17 +124,13 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
         $res = $this->ftp->mkdir('test');
         
         if (PEAR::isError($res)) {
-            $this->setupError = 'Could not create a test directory';
-            $this->ftp        = null;
-            return;
+            $this->markTestSkipped('Could not create a test directory');
         }
         
         $res = $this->ftp->cd('test');
         
         if (PEAR::isError($res)) {
-            $this->setupError = 'Could not change to the test directory';
-            $this->ftp        = null;
-            return;
+            $this->markTestSkipped('Could not change to the test directory');
         }
     }
 
@@ -171,7 +150,6 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
             
             $this->ftpdir     = null;
             $this->ftp        = null;
-            $this->setupError = null;
         }
     }
     
@@ -183,11 +161,6 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
      */
     public function testMkdir()
     {
-        if ($this->ftp == null) {
-            $this->fail('This test requires a working FTP connection. Setup '.
-            'config.php with proper configuration parameters. ('.
-            $this->setupError.')');
-        }
         $this->ftp->mkdir('dir1', false);
         $this->ftp->mkdir('dir1/dir2/dir3/dir4', true);
         $this->assertTrue($this->ftp->cd('dir1/dir2/dir3/dir4'));
@@ -201,11 +174,6 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
      */
     public function testRename()
     {
-        if ($this->ftp == null) {
-            $this->fail('This test requires a working FTP connection. Setup '.
-            'config.php with proper configuration parameters. ('.
-            $this->setupError.')');
-        }
         $this->ftp->put('testfile.dat', 'testfile.dat', FTP_ASCII);
         $this->assertTrue($this->ftp->rename('testfile.dat', 'testfile2.dat'));
     }
@@ -218,11 +186,6 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
      */
     public function testRm()
     {
-        if ($this->ftp == null) {
-            $this->fail('This test requires a working FTP connection. Setup '.
-            'config.php with proper configuration parameters. ('.
-            $this->setupError.')');
-        }
         $list1 = $this->ftp->ls();
         
         $this->ftp->put('testfile.dat', 'testfile.dat', FTP_ASCII);
@@ -235,6 +198,28 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
         
         $this->assertEquals($list1, $list2, 'Directory listing before creation and'.
             ' after creation are not equal');
+        
+        $this->ftp->mkdir('dir1');
+        $this->ftp->cd('dir1');
+        $this->ftp->mkdir('dir1');
+        $this->ftp->mkdir('dir2');
+        $this->ftp->mkdir('dir3');
+        $this->ftp->put('testfile.dat', 'testfile.dat', FTP_ASCII);
+        
+        $this->ftp->cd('..');
+        
+        $this->ftp->rm('dir1/', true, true);
+        
+        $this->ftp->cd('dir1');
+        
+        $list = $this->_getNames($this->ftp->ls());
+        $dirs = array('dir1', 'dir2', 'dir3');
+        
+        sort($list);
+        sort($dirs);
+        
+        $this->assertEquals($list, $dirs, 'The resulting directory listing is not '.
+            'correct');
     }
 
     /**
@@ -245,12 +230,6 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
      */
     public function testPutRecursive()
     {
-        if ($this->ftp == null) {
-            $this->fail('This test requires a working FTP connection. Setup '.
-            'config.php with proper configuration parameters. ('.
-            $this->setupError.')');
-        }
-        
         $tmpdir    = array();
         $tmpfile   = array();
         $tmpdir[]  = System::mktemp(array('-d', 'pearnetftptest'));
@@ -274,7 +253,7 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
             $this->assertFalse(PEAR::isError($dirlist));
           
             $dirlist   = $this->_getNames($dirlist);
-            $dirlistok = array($this->_getLastPart($tmpfile[$i]), '.', '..');
+            $dirlistok = array($this->_getLastPart($tmpfile[$i]));
             if ($i < 2) {
                 $dirlistok[] = $this->_getLastPart($tmpdir[$i+1]);
             }
@@ -286,6 +265,23 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
         }
     }
 
+    /**
+     * Tests functionality of Net_FTP::put()
+     * 
+     * @return void
+     * @see Net_FTP::put()
+     */
+    public function testPut()
+    {
+        $res = $this->ftp->put('testfile.dat', 'testfile.dat', true, FTP_ASCII,
+            NET_FTP_BLOCKING | NET_FTP_NONBLOCKING);
+        $this->assertTrue(PEAR::isError($res), 'No error return while it should');
+        if (PEAR::isError($res)) {
+            $this->assertEquals($res->getCode(), NET_FTP_ERR_BADOPTIONS,
+                'Bad error code returned');
+        }
+    }
+
     /**
      * Tests functionality of Net_FTP::_makeDirPermissions()
      * 
@@ -294,11 +290,6 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
      */
     public function testMakeDirPermissions()
     {
-        if ($this->ftp == null) {
-            $this->fail('This test requires a working FTP connection. Setup '.
-            'config.php with proper configuration parameters. ('.
-            $this->setupError.')');
-        }
         $tests = array(
             '111' => '111',
             '110' => '110',
@@ -324,11 +315,6 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
      */
     public function testSize()
     {
-        if ($this->ftp == null) {
-            $this->fail('This test requires a working FTP connection. Setup '.
-            'config.php with proper configuration parameters. ('.
-            $this->setupError.')');
-        }
         // upload in binary to avoid addition/removal of characters
         $this->ftp->put('testfile.dat', 'testfile.dat', FTP_BINARY);
         $this->assertEquals($this->ftp->size('testfile.dat'),
@@ -345,11 +331,6 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
      */
     public function testExtensions()
     {
-        if ($this->ftp == null) {
-            $this->fail('This test requires a working FTP connection. Setup '.
-            'config.php with proper configuration parameters. ('.
-            $this->setupError.')');
-        }
         $this->ftp->setMode(FTP_ASCII);
         $this->ftp->addExtension(FTP_BINARY, 'tst');
         $this->assertEquals($this->ftp->checkFileExtension('test.tst'), FTP_BINARY);
@@ -367,11 +348,6 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
      */
     public function testGetExtensionsFile()
     {
-        if ($this->ftp == null) {
-            $this->fail('This test requires a working FTP connection. Setup '.
-            'config.php with proper configuration parameters. ('.
-            $this->setupError.')');
-        }
         $res = $this->ftp->getExtensionsFile('extensions.ini');
         $this->assertFalse(PEAR::isError($res), 'Test extensions file could be'.
             'loaded');
@@ -382,6 +358,70 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
         $this->assertEquals($this->ftp->checkFileExtension('test.gif'), FTP_BINARY);
     }
 
+    /**
+     * Tests functionality of Net_FTP::setDirMatcher()
+     * 
+     * @since 1.4.0a1
+     * @return void
+     */
+    public function testSetDirMatcher()
+    {
+        $pattern = '/(?:(d)|.)([rwxt-]+)\s+(\w+)\s+(\w+)\s+(\w+)\s(\w+\s+\d+\s+.+)'.
+            '\s+(.+)/';
+        $map     = array(
+            'is_dir' => 1,
+            'rights' => 2,
+            'user' => 3,
+            'group' => 4,
+            'size' => 5,
+            'date' => 6,
+            'name' => 7,
+        );
+        
+        $res = $this->ftp->setDirMatcher(false, false);
+        $this->assertTrue(PEAR::isError($res)
+            && $res->getCode() == NET_FTP_ERR_ILLEGALPATTERN,
+            'Result of setDirMatcher(boolean, boolean) should give a '.
+            'NET_FTP_ERR_ILLEGALPATTERN error');
+        
+        $res = $this->ftp->setDirMatcher($pattern, false);
+        $this->assertTrue(PEAR::isError($res)
+            && $res->getCode() == NET_FTP_ERR_ILLEGALMAP,
+            'Result of setDirMatcher(string, boolean) should give a '.
+            'NET_FTP_ERR_ILLEGALMAP error');
+        
+        $res = $this->ftp->setDirMatcher($pattern, array('a' => 'b'));
+        $this->assertTrue(PEAR::isError($res)
+            && $res->getCode() == NET_FTP_ERR_ILLEGALMAPVALUE,
+            'The items in a matcher map should only contain numeric values');
+        
+        $res = $this->ftp->setDirMatcher($pattern, $map);
+        $this->assertFalse(PEAR::isError($res),
+            'A valid pattern and map should return no error');
+    }
+    
+    /**
+     * Tests changes made to fix bug #9611
+     * 
+     * @since 1.4.0a1
+     * @return void
+     */
+    public function testCheckRemoteDir()
+    {
+        $res = $this->ftp->_checkRemoteDir('test');
+        $this->assertFalse($res, 'test should not be a directory');
+        
+        $this->ftp->put('testfile.dat', 'test', FTP_BINARY);
+        $res = $this->ftp->_checkRemoteDir('test');
+        $this->assertFalse($res, 'test should not be a directory');
+        
+        $this->ftp->rm('test');
+        
+        $this->ftp->mkdir('test');
+        $res = $this->ftp->_checkRemoteDir('test');
+        $this->assertTrue($res, 'test should be a directory');
+    }
+
     /**
      * Tests _determineOSMatch
      * 
@@ -390,25 +430,21 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
      */
     public function testDetermineOSMatch()
     {
-        if ($this->ftp == null) {
-            $this->fail('This test requires a working FTP connection. Setup '.
-            'config.php with proper configuration parameters. ('.
-            $this->setupError.')');
-        }
         $dirlist = array(
-            'drwxrwsr-x  75 upl.oad  (?).         3008 Oct 30 21:09 ftp1',
+            'drwxrwsr-x  75 upload.  (?).         3008 Oct 30 21:09 ftp1',
         );
         
         $res = $this->ftp->_determineOSMatch($dirlist);
         $this->assertFalse(PEAR::isError($res),
             'The directory listing should be recognized');
+        
         $this->assertEquals($res['pattern'],
             $this->ftp->_ls_match['unix']['pattern'],
             'The input should be parsed by the unix pattern');
     }
     
     /**
-     * Return all name keys in the elements of an array
+     * Return all name keys in the elements of an array, leaving out . and ..
      *
      * @param array $in Multidimensional array
      *
@@ -418,7 +454,9 @@ class Net_FTPTest extends PHPUnit_Framework_TestCase
     {
         $return = array();
         foreach ($in as $v) {
-            $return[] = $v['name'];
+            if ($v['name'] != '.' && $v['name'] != '..') {
+                $return[] = $v['name'];
+            }
         }
         return $return;
     }
